
from collections import OrderedDict
from cStringIO import StringIO
from grp import getgrnam
from pprint import pformat
from pwd import getpwnam
from socket import gethostbyaddr
from subprocess import check_output, STDOUT, CalledProcessError
from tabulate import tabulate
import errno
import json
import logging
import logging.handlers
import os
import re
import sys

# Used to support ifupdown2's glob syntax
sys.path.append('/usr/share/ifupdown2/')
import ifupdownaddons.modulebase
modbase = ifupdownaddons.modulebase.moduleBase()

log = logging.getLogger('netd')

pending_commands_file = '/var/run/nclu/pending_commands.txt'


class Capturing(list):

    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        sys.stdout = self._stdout


def files_match(fileA, fileB):
    if not os.path.isfile(fileA) and not os.path.isfile(fileB):
        return True

    if not os.path.isfile(fileA) or not os.path.isfile(fileB):
        return False

    fileA_lines = []
    fileB_lines = []

    with open(fileA, 'r') as fh:
        fileA_lines = fh.readlines()

    with open(fileB, 'r') as fh:
        fileB_lines = fh.readlines()

    return fileA_lines == fileB_lines


def print_diff(fileA, fileB, color_diffs):
    """
    Print the diff between fileA vs. fileB with colored output
    Return True if we printed anything
    """
    if not os.path.isfile(fileA) or not os.path.isfile(fileB):
        return False

    try:
        check_output(['/usr/bin/diff',
                      '-U', '10',
                      '--ignore-blank-lines',
                      '--ignore-space-change',
                      fileA,
                      fileB])
        return False
    except CalledProcessError as e:

        # Print the deletes in red and the adds in green
        for line in e.output.splitlines():
            if color_diffs:
                if line.startswith('---') or line.startswith('+++'):
                    print "\033[94m%s\033[0m" % line
                elif line.startswith('-'):
                    print "\033[91m%s\033[0m" % line
                elif line.startswith('+'):
                    print "\033[92m%s\033[0m" % line
                else:
                    print line
            else:
                print line
        return True


def chown_chgroup(filename, user, group):
    quagga_uid = getpwnam(user).pw_uid
    quagga_gid = getgrnam(group).gr_gid
    os.chown(filename, quagga_uid, quagga_gid)


def file_contains(filename, target_line):
    target_line = target_line.strip()

    if os.path.isfile(filename):
        with open(filename, 'r') as fh:
            for line in fh.readlines():
                line = line.strip()
                if line == target_line:
                    return True
    return False


def ifname_is_valid(ifname):
    """
    Enforce the rules defined in the kernel's dev_valid_name()
    """
    IF_NAMESIZE = 16
    illegal_characters = ('/', ' ', '%', '*')

    if ifname == '.' or ifname == ':':
        print "'%s' is an invalid interface name (cannot use '.' or '..')" % ifname
        return False

    for character in illegal_characters:
        if character in ifname:
            print "'%s' is an invalid interface name (cannot use '%s')" % (ifname, '\', \''.join(illegal_characters))
            return False

    if len(ifname) > IF_NAMESIZE:
        print "'%s' is an invalid interface name (more than %d characters)" % (ifname, IF_NAMESIZE)
        return False

    return True


def ifname_is_globish(ifname):
    """
    This assumes that ifname_is_glob() has already been called for ifname
    and has verified that ifname is not a glob

    Return True if the ifname contains characters that one might use when
    attempting to enter a glob
    """
    globish_characters = (',', '[', ']', '(', ')', '%', '*')

    glob_msg = """
'%s' looks sort of like a glob but does not use valid glob syntax.
globs use '-' to specify a range of interfaces.

Examples:
    swp1-4
    swp1-4,swp8,swp9
    swp1-4,8,9
    swp1-4,swp8-9
    swp1,swp8,swp9
    swp1,8,9
    swp1,swp8-9

""" % ifname

    re_glob_fakeout = re.search('^\w+0\d+$', ifname)
    if re_glob_fakeout:
        return False

    re_glob_fakeout = re.search('^\w+0\d+-0\d+$', ifname)
    if re_glob_fakeout:
        return False

    re_swp_swp = re.search('^swp\d+-\w+\d+$', ifname)
    if re_swp_swp:
        print glob_msg
        return True

    for character in globish_characters:
        if character in ifname:
            print glob_msg
            return True

    return False

def ifname_is_glob(value):

    if not value:
        return False

    # server01 this is not a glob
    re_glob_fakeout = re.search('^\w+0\d+$', value)
    if re_glob_fakeout:
        return False


    re_glob_fakeout = re.search('^\w+0\d+-0\d+$', value)
    if re_glob_fakeout:
        return False

    # These are globs:
    #  swp1-5,swp8
    #  swp1-5,8
    #  swp1-5,swp7-8
    #  swp1-5,swp7
    #
    # This is not a glob
    #  server01
    if ',' in value:
        glob_parts = value.split(',')
        base_swp = glob_parts[0]

        if modbase.ifname_is_glob(base_swp):
            for x in glob_parts[1:]:
                re_swp = re.search('[a-zA-Z]+\d+', x)
                re_number_number = re.search('\d+-\d+', x)

                if modbase.ifname_is_glob(x) or x.isdigit() or re_swp or re_number_number:
                    pass
                else:
                    return False

            return True

        if ifname_is_valid(base_swp):
            for x in glob_parts[1:]:
                re_swp = re.search('[a-zA-Z]+\d+', x)
                re_number_number = re.search('\d+-\d+', x)

                if modbase.ifname_is_glob(x) or x.isdigit() or re_swp or re_number_number:
                    pass
                else:
                    return False

            return True

    # swp1-5
    elif modbase.ifname_is_glob(value):
        return True

    return False


def ifname_expand_glob(value):
    ifnames = []

    # swp1-5,swp8,swp9
    # swp1-5,8,9
    # swp1-5,swp8-9
    # swp1,swp8,swp9
    # swp1,8,9
    # swp1,swp8-9
    if ',' in value:
        glob_parts = value.split(',')
        base_swp = glob_parts[0]

        if modbase.ifname_is_glob(base_swp):

            # base_swp is a glob so expand it and take the first interface
            # as our new base_swp...when we do this strip
            for ifname in modbase.parse_glob('foobarblah', base_swp):
                ifnames.append(ifname)
            base_swp = ifnames[0]

        elif ifname_is_valid(base_swp):
            ifnames.append(base_swp)
        else:
            base_swp = None

        if base_swp:
            base_swp = base_swp.rstrip('1234567890')

            for x in glob_parts[1:]:
                re_swp = re.search('^[a-zA-Z]+\d+$', x)
                re_number_number = re.search('(\d+)-(\d+)', x)

                if modbase.ifname_is_glob(x):
                    ifnames.extend(modbase.parse_glob('foobarblah', x))

                elif x.isdigit():
                    ifnames.append("%s%s" % (base_swp, x))

                elif re_swp:
                    ifnames.append(x)

                elif re_number_number:
                    min_x = int(re_number_number.group(1))
                    max_x = int(re_number_number.group(2)) + 1

                    for y in xrange(min_x, max_x):
                        ifnames.append("%s%d" % (base_swp, y))
                else:
                    log.info("unknown glob component '%s' in ifname_expand_glob()" % x)

    # swp1-5
    elif modbase.ifname_is_glob(value):
        for ifname in modbase.parse_glob('foobarblah', value):
            ifnames.append(ifname)

    return ifnames


def ifnames_to_glob(ifnames):

    if len(ifnames) == 1:
        return ifnames[0]

    ifnumbers = []

    # First make sure all interfaces begin with the same prefix and that none
    # of these are sub-interfaces (no . in the name)
    prev_prefix = None
    for ifname in ifnames:
        # server01....we cannot change this to server1
        re_glob_fakeout = re.search('^\w+0\d+$', ifname)
        re_glob_fakeout2 = re.search('^\w+0\d+-0\d+$', ifname)

        re_ifname = re.search('^(.*?)(\d.*)$', ifname)

        if re_glob_fakeout or re_glob_fakeout2:
            pass

        # swp1, swp2, etc
        elif re_ifname:
            prefix = re_ifname.group(1)
            suffix = re_ifname.group(2)

            # not supported for now
            if '.' in suffix:
                return None

            if not suffix.isdigit():
                return None

            if prev_prefix is not None and prefix != prev_prefix:
                return None

            ifnumbers.append(int(suffix))
            prev_prefix = prefix

    if prev_prefix and ifnumbers:
        return prev_prefix + numbers_to_glob(ifnumbers)
    return None


def numbers_to_glob(ids):
    """
    Given:
     (3,4,7,10,11,12,17,22,4001,4002,4003,7777,8000,8001,8002,8003,8004)

    Return:
     3-4,7,10-12,17,22,4001-4003,7777,8000-8004
    """
    ids_final = []

    for _id in ids:
        if isinstance(_id, int):
            ids_final.append(_id)
        elif _id.isdigit():
            ids_final.append(int(_id))
        else:
            re_number_range = re.match('(\d+)-(\d+)', _id)
            if re_number_range:
                start = int(re_number_range.group(1))
                end = int(re_number_range.group(2)) + 1
                ids_final.extend(range(start, end))
    ids = sorted(list(set(ids_final)))

    min_id = ids[0]
    max_id = ids[-1]
    limiters = []

    if len(ids) == (max_id - min_id + 1):
        if min_id == max_id:
            limiters.append("%d" % min_id)
        else:
            limiters.append("%d-%d" % (min_id, max_id))
    else:
        streak_start_x = 0
        streak_end_x = 0
        len_ids = len(ids)

        for x in range(len_ids):

            # The first number in the list
            if x == 0:
                current = ids[x]
                next_ids = ids[x + 1]

                if current + 1 == next_ids:
                    streak_start_x = current
                else:
                    limiters.append("%d" % current)

            # The last number in the list
            elif x == len_ids - 1:
                prev_ids = ids[x - 1]
                current = ids[x]

                if prev_ids + 1 == current:
                    streak_end_x = current
                    limiters.append("%d-%d" % (streak_start_x, streak_end_x))
                else:
                    limiters.append("%d" % current)

            # Any number in the middle of the list
            else:
                prev_ids = ids[x - 1]
                current = ids[x]
                next_ids = ids[x + 1]

                # In the middle of a streak...do nothing
                if prev_ids + 1 == current and current + 1 == next_ids:
                    pass

                # The end of a streak
                elif prev_ids + 1 == current:
                    streak_end_x = current
                    limiters.append("%d-%d" % (streak_start_x, streak_end_x))
                    streak_start_x = 0
                    streak_end_x = 0

                # The beginning of a streak
                elif current + 1 == next_ids:
                    streak_start_x = current

                # A loner
                else:
                    limiters.append("%d" % current)

    return ','.join(limiters)


def glob_to_numbers(glob):

    try:
        glob = glob.strip().replace(',', ' ').split()

    # will see an AttributeError on strip() if glob wasn't a string, unicode, etc
    except AttributeError:
        pass

    glob = list(set(glob))
    numbers = []

    for x in glob:
        if x.isdigit():
            numbers.append(int(x))
        else:
            re_range = re.match('^(\d+)-(\d+)$', x)

            if re_range:
                min_range = int(re_range.group(1))
                max_range = int(re_range.group(2))

                if max_range >= min_range:
                    numbers.extend(range(min_range, max_range + 1))
                else:
                    print "%s is an invalid number range" % x
                    log.error("globs must consist of numbers or ranges, %s is neither. We were given glob '%s'" % (x, glob))
                    return []
            else:
                print "%s is an invalid number range" % x
                log.error("globs must consist of numbers or ranges, %s is neither. We were given glob '%s'" % (x, glob))
                return []

    return sorted(list(set(numbers)))


def is_number_glob(glob):
    """
    Return True if glob is a numeric glob such as 1-5,10,13-15,20
    glob can be a string or a list
    """
    if glob is None:
        return False

    try:
        glob = glob.strip().replace(',', ' ').split()
    except Exception:
        pass

    found_digit = False

    for x in glob:
        re_range = re.match('^\d+-\d+$', x)

        if x.isdigit() or re_range:
            found_digit = True
        else:
            return False

    return found_digit


def get_lldp():
    lldp = {}

    if os.path.exists('/usr/sbin/lldpctl'):
        try:
            output = check_output(['/usr/sbin/lldpctl', '-f', 'json'])
            lldp_original = json.loads(output)
            for value in lldp_original['lldp'][0].get('interface'):
                # log.info("lldp value %s" % pformat(value))
                lldp[value['name']] = value['chassis'][0]['name'][0]['value']

        except Exception as e:
            log.exception(e)

    return lldp


def tabulate_pending_add_del_commands():
    header = ['User', 'Timestamp', 'Command']
    data = []

    if os.path.exists(pending_commands_file):
        with open(pending_commands_file, 'r') as fh:
            for line in fh.readlines():
                (user, timestamp, command)= line.strip().split('\t')
                data.append((user, timestamp, command))

    return tabulate(data, header)


def print_pending_add_del_commands():
    print "Please run 'net abort' and use net to re-issue your changes."

    if os.path.exists(pending_commands_file):
        print "\nThe following is a list of all add/del"
        print "commands since the last 'net commit'"
        print "======================================"

        with open(pending_commands_file, 'r') as fh:
            for line in fh.readlines():
                (user, timestamp, command)= line.strip().split('\t')
                print command
        print ''


def globify_interface_commands(init_commands):
    """
    Given:
    net add int swp1 mtu 1500
    net add int swp2 mtu 1500
    net add int swp3 mtu 1500

    Return:
    net add int swp1-3 mtu 1500
    """
    final_commands = []
    globbed_commands = []
    non_globbed_commands = []
    commands_by_id = OrderedDict()

    for command in init_commands:
        command_list = command.split()

        # net add bridge
        if len(command_list) <= 3:
            final_commands.append(command)
            continue

        add_del = command_list[1]
        iface_type = command_list[2]

        # net add vlan 10-20
        if iface_type == 'vlan':
            final_commands.append(command)
            continue

        ifname = command_list[3]
        command_id = (add_del, iface_type, ' '.join(command_list[4:]))

        if command_id not in commands_by_id:
            commands_by_id[command_id] = []
        commands_by_id[command_id].append(ifname)

    # We must configure the bond-slaves, vlan-interface, etc statements first
    for bond_slaves_etc_first in (True, False):
        for (command_id, ifnames) in commands_by_id.iteritems():
            (add_del, iface_type, keyword_value) = command_id

            if (keyword_value.startswith('bond-slaves') or
                keyword_value.startswith('vlan-interface') or
                keyword_value.startswith('vni')):
                if not bond_slaves_etc_first:
                    continue
            else:
                if bond_slaves_etc_first:
                    continue

            if len(ifnames) > 1:
                ifnames_glob = ifnames_to_glob(ifnames)
            else:
                ifnames_glob = None

            if ifnames_glob:
                globbed_commands.append("net %s %s %s %s" % (add_del, iface_type, ifnames_glob, keyword_value))
            else:
                for ifname in ifnames:
                    non_globbed_commands.append("net %s %s %s %s" % (add_del, iface_type, ifname, keyword_value))

    final_commands.extend(globbed_commands)
    final_commands.extend(non_globbed_commands)
    return final_commands


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def seconds_to_hhmmss(secs):

    if not secs:
        return 'never'

    mins, secs = divmod(secs, 60)
    hours, mins = divmod(mins, 60)
    days, hours = divmod(hours, 24)

    if days:
        return '%d day%s, %02d:%02d:%02d' % (days, 's' if days > 1 else '',  hours, mins, secs)
    else:
        return '%02d:%02d:%02d' % (hours, mins, secs)


def get_hostname(ip):
    try:
        hostname = gethostbyaddr(ip)
        return " (%s)" % hostname[0]
    except Exception:
        return ""


def file_set_key_value(filename, key, value, comment_out=False):
    """
    For config files that follow a key=value notation
    ip=10.1.1.1

    This function provides a way to alter the value for a key
    """

    if os.path.exists(filename):
        final = []

        with open(filename, 'r') as fh:
            for line in fh.readlines():
                line = line.strip()
                re_key = re.search('^\s*%s(\s*)=(\s*)' % key, line)
                re_key_commented = re.search('^\s*#\s*%s(\s*)=(\s*)' % key, line)

                if re_key:
                    leading_spaces = re_key.group(1)
                    trailing_spaces = re_key.group(2)
                    final.append("%s%s=%s%s\n" % (key, leading_spaces, trailing_spaces, value))
                elif re_key_commented:
                    leading_spaces = re_key_commented.group(1)
                    trailing_spaces = re_key_commented.group(2)
                    final.append("%s%s=%s%s\n" % (key, leading_spaces, trailing_spaces, value))
                else:
                    final.append(line + '\n')

        with open(filename, 'w') as fh:
            for line in final:
                fh.write(line)

    else:
        log.info("Could not set %s to %s in %s, the file does not exist" % (key, value, filename))
