#!/usr/bin/env python

"""
Usage:

    # LNV: Lightweight Network Virtualization
    # LNV configuration
    lnv (add|del) lnv service-node anycast-ip <ipv4>
    lnv (add|del) lnv service-node source <ipv4>
    lnv (add|del) lnv service-node peers <wildcard>

    # LNV show commands
    lnv show lnv

Options:

    anycast-ip   : An anycast IPv4 address
    lnv          : Lightweight Network Virtualization
    peers        : A peer, neighbor, etc
    service-node : LNV service-node
    source       : Source IPv4 address
    vtep         : VXLAN tunnel end point

"""

from collections import OrderedDict
from network_docopt import NetworkDocopt, sort_for_humans
from nclu import print_diff, file_set_key_value, file_set_key_value, mkdir_p, files_match, get_lldp
from tabulate import tabulate
from itertools import izip_longest
from pprint import pformat
from subprocess import check_output, STDOUT, CalledProcessError, call
import json
import logging
import logging.handlers
import os
import re
import shutil
import shlex
import sys

log = logging.getLogger('netd')
var_run_nclu = '/var/run/nclu/lnv/'
vxrd_conf = '/etc/vxrd.conf'
vxsnd_conf = '/etc/vxsnd.conf'

# In the initial implementation of this wrapper we had to modify both vxrd.conf
# and vxsnd.conf.  We no longer need to modify vxrd.conf (we use keywords in
# ifupdown2 to configure this today) but I am leaving the references to
# vxrd.conf in the code in case we ever wrapper modifying some of the settings
# in that file.
execlist = '/var/run/nclu/lnv/execlist'
vxrd_conf_scratchpad = '/var/run/nclu/lnv/vxrd.conf'
vxsnd_conf_scratchpad = '/var/run/nclu/lnv/vxsnd.conf'


def ifaces_in_vlan(bridge_vlan_show, target_vlan):
    ifaces = []
    for (ifname, value) in bridge_vlan_show.iteritems():
        # value is a list of dictionaries
        for entry in value:
            vlan = entry.get('vlan')

            if vlan == target_vlan:
                ifaces.append(ifname)

    return sort_for_humans(list(set(ifaces)))


def show_lnv():
    """
    Pull data from various places to print a summary of the LNV state on this node
    """

    # Are we a vtep?
    try:
        check_output(['/bin/systemctl', 'is-active', 'vxsnd.service'])
        is_service_node = True
    except CalledProcessError:
        is_service_node = False

    # Are we a service node?
    try:
        check_output(['/bin/systemctl', 'is-active', 'vxrd.service'])
        is_vtep = True
    except CalledProcessError:
        is_vtep = False

    if is_service_node:
        '''
        {"peers": [], "version": 3, "anycast": "6.0.0.9", "ctrl_port": 10001, "data_port": 4789, "local": "6.0.0.9"}

        Protocol version: 3
        Local address      Anycast address    Peer Addresses
        =============      ===============    ==============
        6.0.0.9            6.0.0.9

        Data (UDP) Port                       Control (UDP) Port
        ==============                        =================
        4789                                  10001

        '''
        try:
            output = check_output(['/usr/bin/vxsndctl', 'show', 'detail', '-j'])
        except CalledProcessError:
            output = "{}"
        vxsnd_show = json.loads(output)


        '''
        {"10": [["6.0.0.12", "76", "89254930658124"], ["6.0.0.11", "86", "87057984938359"]]}

        VNI         Address         Ageout
        ===         =======         ======
        10          6.0.0.11        85
        10          6.0.0.12        75

        '''
        try:
            output = check_output(['/usr/bin/vxsndctl', 'fdb', '-j'])
        except CalledProcessError:
            output = "{}"
        vxsnd_fdb = json.loads(output)

        output = []
        output.append("")
        output.append("  LNV Role           : Service Node")
        output.append("  Version            : %s" % vxsnd_show.get('version'))
        output.append("  Local IP           : %s" % vxsnd_show.get('local'))
        output.append("  Anycast IP         : %s" % vxsnd_show.get('anycast'))
        output.append("  UDP Data Port      : %s" % vxsnd_show.get('data_port'))
        output.append("  UDP Control Port   : %s" % vxsnd_show.get('ctrl_port'))
        output.append("  Service Node Peers : %s" % ', '.join(vxsnd_show.get('peers')))
        output.append("")

        header = ('VNI', 'VTEP', 'Ageout')

        data = []
        for (vni, values) in  vxsnd_fdb.iteritems():
            for (ip, ageout, timeout) in values:
                data.append((vni, ip, ageout))

        final_data = []
        prev_vni = None
        for (vni, ip, ageout) in sorted(data):
            if prev_vni is None or prev_vni != vni:
                final_data.append((vni, ip, ageout))
            else:
                final_data.append(('', ip, ageout))
            prev_vni = vni

        for line in tabulate(final_data, header).splitlines():
            output.append('  ' + line)
        output.append("")
        output.append("")

        print "\n".join(output)

    if is_vtep:
        '''
        {"10": {"dev_name": "vni-10", "svcnodeip": "6.0.0.9", "hrep_addrs": ["6.0.0.12"], "localip": "6.0.0.11"}}

        VNI         Local Addr      Svc Node
        ===         ==========      ========
        10          6.0.0.11        6.0.0.9
        '''
        try:
            # This takes ~350ms but we have to have this data
            output = check_output(['/usr/bin/vxrdctl', 'vxlans', '-j'])
        except CalledProcessError:
            output = "{}"
        vxrd_vxlans = json.loads(output)


        '''
        {"10": ["6.0.0.12", "6.0.0.11"]}

        VNI         Peer Addrs
        ===         ==========
        10          6.0.0.11, 6.0.0.12

        '''
        # No new info here d what we already learned via 'vxrdctl vxlans'
        # output = check_output(['/usr/bin/vxrdctl', 'peers', '-j'])
        # peers = json.loads(output)


        '''
        {"src_ip": "6.0.0.11", "snd_ip": "6.0.0.9", "version": 3, "head_rep": true}

        Protocol version: 3
        Local address      Svcnode address    Head rep.
        =============      ===============    =========
        6.0.0.11           6.0.0.9            Enabled

        '''
        # Here we learn the version # and if headend replication is enabled
        try:
            # This takes ~200ms but we have to have this data
            output = check_output(['/usr/bin/vxrdctl', 'show', '-j'])
        except CalledProcessError:
            output = "{}"
        vxrd_show = json.loads(output)

        try:
            output = check_output(['bridge', '-j', 'vlan', 'show'])
        except CalledProcessError:
            output = "{}"
        bridge_show = json.loads(output)

        try:
            output = check_output(['/sbin/brctl', 'show'])
        except CalledProcessError:
            output = ""

        brctl_show = {}
        for line in output.splitlines():
            if line.startswith('bridge name'):
                continue

            columns = line.replace('\t', ' ').split()

            '''
            bridge name bridge id           STP enabled interfaces
            br-10       8000.002590b201b9   yes         swp3
                                                        swp4
                                                        vni-10
            '''
            if len(columns) == 4:
                bridge_name = columns[0]
                # bridge_id = columns[1]
                # step = columns[2]
                iface = columns[3]
                brctl_show[bridge_name] = [iface, ]
            else:
                brctl_show[bridge_name].append(columns[0])

        output = []
        output.append("")
        output.append("  LNV Role            : VTEP")
        output.append("  Version             : %s" % vxrd_show.get('version'))
        output.append("  Headend Replication : %s" % vxrd_show.get('head_rep'))
        output.append("")

        lldp = get_lldp()

        header = ('VLAN', 'VNI', 'Local IP', 'Service Node IP', 'VLAN Members', 'VTEP Peers')
        data = []

        for (vni, value) in vxrd_vxlans.iteritems():
            ifaces = []
            vni_bridge = None
            vni_vlan = None
            vni_iface = value['dev_name']
            vni_local_ip = value['localip']
            vni_svcnode_ip = value['svcnodeip']

            vni_bridge_show = bridge_show.get(vni_iface)

            if vni_bridge_show:
                vni_vlan = vni_bridge_show[0].get('vlan')
                ifaces = ifaces_in_vlan(bridge_show, vni_vlan)

            # Which bridge is the vni interface in?
            #for (bridge, bridge_ifaces) in brctl_show.iteritems():
            #    if vni_iface in bridge_ifaces:
            #        vni_bridge = bridge
            #        break

            peers = sort_for_humans(value['hrep_addrs'])

            for (iface, peer) in izip_longest(ifaces, peers):
                if iface:
                    lldp_peer = lldp.get(iface)

                    if lldp_peer:
                        iface = "%s (%s)" % (iface, lldp_peer)

                data.append((vni_vlan, vni, vni_local_ip, vni_svcnode_ip, iface, peer))

        final_data = []
        prev_vni_vlan = None
        for (vni_vlan, vni, vni_local_ip, vni_svcnode_ip, iface, peer) in sorted(data):
            if prev_vni_vlan is None or prev_vni_vlan != vni_vlan:
                final_data.append((vni_vlan, vni, vni_local_ip, vni_svcnode_ip, iface, peer))
            else:
                final_data.append(('', '', '', '', iface, peer))

            prev_vni_vlan = vni_vlan

        for line in tabulate(final_data, header).splitlines():
            output.append('  ' + line)
        output.append("")
        print "\n".join(output)


def get_vxsnd_conf(filename):
    svcnode_ip = None
    src_ip = None
    svcnode_peers = None

    if os.path.exists(filename):
        with open(filename, 'r') as fh:
            for line in fh.readlines():

                if line.startswith('svcnode_ip'):
                    try:
                        svcnode_ip = line.split('=')[1].strip()
                    except Exception:
                        svcnode_ip = None

                elif line.startswith('src_ip'):
                    try:
                        src_ip = line.split('=')[1].strip()
                    except Exception:
                        src_ip = None

                elif line.startswith('svcnode_peers'):
                    try:
                        svcnode_peers = line.split('=')[1].strip()
                    except Exception:
                        svcnode_peers = None

    return (svcnode_ip, src_ip, svcnode_peers)


def vxsnd_conf_modified(filename):
    """
    Return True if any of the following are True
    - svcnode_ip is not 0.0.0.0
    - src_ip is not 0.0.0.0
    - svcnode_peers is not ""
    """
    if os.path.exists(filename):
        with open(filename, 'r') as fh:
            for line in fh.readlines():
                line = line.replace(' ', '').strip()

                if line.startswith('svcnode_ip'):
                    if line != 'svcnode_ip=0.0.0.0':
                        return True

                elif line.startswith('src_ip'):
                    if line != 'src_ip=0.0.0.0':
                        return True

                elif line.startswith('svcnode_peers'):
                    if line != 'svcnode_peers=':
                        return True

    return False


class ConfigWrapper(object):

    def __init__(self, show_linux_command, color_diffs):
        self.show_linux_command = show_linux_command
        self.color_diffs = color_diffs

        if not os.path.exists(var_run_nclu):
            mkdir_p(var_run_nclu)

    def get_managed_files(self):
        files = []
        files.append(vxrd_conf)
        files.append(vxsnd_conf)
        return files

    def get_service_cmd_for_file(self, filename):
        cmds = []

        if filename == vxrd_conf:
            cmds.append(('/bin/systemctl', 'reset-failed', 'vxrd.service'))
            cmds.append(('/bin/systemctl', 'restart', 'vxrd.service'))

        elif filename == vxsnd_conf:
            cmds.append(('/bin/systemctl', 'reset-failed', 'vxsnd.service'))
            cmds.append(('/bin/systemctl', 'restart', 'vxsnd.service'))

        return cmds

    def abort_pending(self):
        if os.path.exists(var_run_nclu):
            for filename in os.listdir(var_run_nclu):
                full_filename = os.path.join(var_run_nclu, filename)

                if os.path.isdir(full_filename):
                    rmtree(full_filename)
                else:
                    os.remove(full_filename)

    def commit_pending(self, verbose):
        files_modified = []

        if os.path.exists(vxrd_conf_scratchpad) and not files_match(vxrd_conf, vxrd_conf_scratchpad):
            shutil.move(vxrd_conf_scratchpad, vxrd_conf)
            files_modified.append(vxrd_conf)

        if os.path.exists(vxsnd_conf_scratchpad) and not files_match(vxsnd_conf, vxsnd_conf_scratchpad):
            shutil.move(vxsnd_conf_scratchpad, vxsnd_conf)
            files_modified.append(vxsnd_conf)


        if os.path.exists(execlist):
            with open(execlist, 'r') as fh:
                for line in fh.readlines():
                    cmd = shlex.split(line.strip())

                    if self.show_linux_command:
                        print "exec: %s" % ' '.join(cmd)

                    try:
                        output = check_output(cmd, stderr=STDOUT)
                        if verbose:
                            print output
                    except CalledProcessError as e:
                        print "\nWARNING: '%s' failed due to\n%s\n" % (' '.join(cmd), e)

            os.remove(execlist)

        return files_modified

    def show_pending(self, print_errors=True):
        diff_pending = False

        if os.path.exists(vxrd_conf_scratchpad):
            if print_diff(vxrd_conf, vxrd_conf_scratchpad, self.color_diffs):
                diff_pending = True

        if os.path.exists(vxsnd_conf_scratchpad):
            if print_diff(vxsnd_conf, vxsnd_conf_scratchpad, self.color_diffs):
                diff_pending = True

        return diff_pending

    def show_config_files(self, user_may_edit):

        # for filename in (vxrd_conf, vxsnd_conf):
        for filename in (vxsnd_conf, ):
            if os.path.exists(filename) and os.path.getsize(filename) > 0:

                # If we haven't modified the default config file then do not display it
                if filename == vxsnd_conf and not vxsnd_conf_modified(vxsnd_conf):
                    continue

                print filename
                print '=' * len(filename)

                with open(filename, 'r') as fh:
                    print ''.join(fh.readlines())
                print ''

    def show_config_commands(self, user_may_edit):
        commands = []
        (svcnode_ip, src_ip, svcnode_peers) = get_vxsnd_conf(vxsnd_conf)

        if svcnode_ip and svcnode_ip != '0.0.0.0':
            commands.append('net add lnv service-node anycast-ip %s' % svcnode_ip)

        if src_ip and src_ip != '0.0.0.0':
            commands.append('net add lnv service-node source %s' % src_ip)

        if svcnode_peers:
            commands.append('net add lnv service-node peers %s' % svcnode_peers)

        return (commands, [])

    def show_config_summary(self, user_may_edit, summary):
        (svcnode_ip, src_ip, svcnode_peers) = get_vxsnd_conf(vxsnd_conf)

        if ((svcnode_ip and svcnode_ip != '0.0.0.0') or
            (src_ip and src_ip != '0.0.0.0') or
            svcnode_peers):
            summary["lnv"] = OrderedDict()
            summary["lnv"]["service-node"] = OrderedDict()

            if svcnode_ip and svcnode_ip != '0.0.0.0':
                summary["lnv"]["service-node"]["anycast-ip"] = svcnode_ip

            if src_ip and src_ip != '0.0.0.0':
                summary["lnv"]["service-node"]["source"] = src_ip

            if svcnode_peers:
                summary["lnv"]["service-node"]["peers"] = svcnode_peers

    def eval_command_line_args(self, cli):
        args = cli.args
        # log.info("ARGS:\n%s" % pformat(args))
        command_ok = True

        if cli.args.get('add'):

            # lnv (add|del) lnv service-node anycast-ip <ipv4>
            # lnv (add|del) lnv service-node source <ipv4>
            # lnv (add|del) lnv service-node peers <text>
            if args.get('service-node'):
                if not os.path.exists(vxsnd_conf_scratchpad):
                    shutil.copyfile(vxsnd_conf,  vxsnd_conf_scratchpad)

                if args.get('anycast-ip'):
                    anycast_ip = args.get('<ipv4>')
                    file_set_key_value(vxsnd_conf_scratchpad, 'svcnode_ip', anycast_ip)
                elif args.get('source'):
                    source_ip = args.get('<ipv4>')
                    file_set_key_value(vxsnd_conf_scratchpad, 'src_ip', source_ip)
                elif args.get('peers'):
                    peers = ' '.join(args.get('<wildcard>').replace(',', ' ').strip().split())
                    file_set_key_value(vxsnd_conf_scratchpad, 'svcnode_peers', peers)

                with open(execlist, 'w') as fh:
                    fh.write('/bin/systemctl enable vxsnd.service\n')
                    fh.write('/bin/systemctl reset-failed vxsnd.service\n')
                    fh.write('/bin/systemctl restart vxsnd.service\n')

        elif cli.args.get('del'):

            if args.get('service-node'):
                if args.get('anycast-ip'):
                    file_set_key_value(vxsnd_conf_scratchpad, 'svcnode_ip', '0.0.0.0')
                elif args.get('source'):
                    file_set_key_value(vxsnd_conf_scratchpad, 'src_ip', '0.0.0.0')
                elif args.get('peers'):
                    file_set_key_value(vxsnd_conf_scratchpad, 'svcnode_peers', '')

            if not vxsnd_conf_modified(vxsnd_conf_scratchpad):
                with open(execlist, 'w') as fh:
                    fh.write('/bin/systemctl disable vxsnd.service\n')
                    fh.write('/bin/systemctl stop vxsnd.service\n')

        elif args.get('show'):
            show_lnv()

        return command_ok
