#!/usr/bin/env python

"""
Usage:

    # Linux Basics
    netmisc add hostname <text>
    netmisc show hostname
    netmisc show version

Options:
    help  : show this screen and exit
    dhcp  : Dynamic Host Configuration Protocol
    relay : Relay Agent
"""

from collections import OrderedDict
from nclu import (
    files_match,
    glob_to_numbers,
    ifname_expand_glob,
    ifname_is_glob,
    ifnames_to_glob,
    is_number_glob,
    numbers_to_glob,
    print_diff
)

from network_docopt import NetworkDocopt, get_network_docopt_info, sort_for_humans, sort_ifaces
from pprint import pformat, pprint
from shutil import copyfile
from subprocess import check_output, STDOUT, CalledProcessError
from tabulate import tabulate
import json
import logging
import logging.handlers
import os
import re
import shlex
import shutil
import sys

try:
    from netshowlib import (interface_related, check as vendor_check, \
        ShowInterfaces, CumulusShowInterfaces, \
        ShowNeighbors, CumulusShowNeighbors, \
        ShowSystem, CumulusShowSystem)
    netshow_present = True
except ImportError:
    netshow_present = False

log = logging.getLogger('netd')

isc_dhcp_relay = '/etc/default/isc-dhcp-relay'
isc_dhcp_relay_scratchpad = '/var/run/nclu/netmisc/isc_dhcp_relay'

ports_conf = '/etc/cumulus/ports.conf'
ports_conf_scratchpad = '/var/run/nclu/netmisc/ports_conf'

etc_hosts = '/etc/hosts'
etc_hostname = '/etc/hostname'
etc_hosts_scratchpad = '/var/run/nclu/netmisc/etc_hosts'
etc_hostname_scratchpad = '/var/run/nclu/netmisc/etc_hostname'
execlist = '/var/run/nclu/netmisc/execlist'


def ifname_to_ports_conf_index(ifname):
    re_swp = re.search('^swp(\d+)', ifname)
    if re_swp:
        return int(re_swp.group(1))
    return None


def ports_conf_edit(args):
    command_ok = True

    if not os.path.exists(ports_conf_scratchpad):
        if not os.path.exists(ports_conf):
            print 'ERROR: %s does not exist' % ports_conf
            return
        copyfile(ports_conf, ports_conf_scratchpad)

    lines = []
    with open(ports_conf_scratchpad, 'r') as fh:
        for line in fh.readlines():
            line = line.strip()
            lines.append(line)

    # netmisc (add|del) ports <number-range-list> speed (4x10G|4x25G|2x50G|10G|40G|50G|100G|disabled)
    for new_speed in ('4x10G', '4x25G', '2x50G', '1G', '10G', '40G', '50G', '100G', 'disabled'):
        if args.get(new_speed):
            break
    else:
        # This should never happen
        print "ERROR: speed for port was not specified"
        return False

    ports_to_edit = []
    ifname = args.get('<interface>')

    if ifname_is_glob(str(ifname)):
        for x in ifname_expand_glob(str(ifname)):
            ports_conf_index = ifname_to_ports_conf_index(x)

            if ports_conf_index:
                ports_to_edit.append(ports_conf_index)
    else:
        ports_conf_index = ifname_to_ports_conf_index(ifname)

        if ports_conf_index:
            ports_to_edit.append(ports_conf_index)

    with open(ports_conf_scratchpad, 'w') as fh:
        for line in lines:
            re_line = re.search('^(\d+)\s*=\s*(\w+)$', line)

            if re_line:
                port_number = int(re_line.group(1))
                port_speed = re_line.group(2)

                if port_number in ports_to_edit:
                    if port_speed == new_speed:
                        print "'%s' configuration already has '%s'" % (port_number, new_speed)
                    fh.write("%s=%s\n" % (port_number, new_speed))
                else:
                    fh.write("%s=%s\n" % (port_number, port_speed))
            else:
                fh.write(line + '\n')

    return command_ok


def dhcp_relay_modified():
    """
    Return True if SERVERS or INTF_CMD are not ""
    """
    if os.path.exists(isc_dhcp_relay):
        with open(isc_dhcp_relay, 'r') as fh:
            for line in fh.readlines():
                line = line.strip()

                if line.startswith('SERVERS'):
                    if line.replace(' ', '') != 'SERVERS=""':
                        return True

                elif line.startswith('INTF_CMD'):
                    if line.replace(' ', '') != 'INTF_CMD=""':
                        return True

    return False


def dhcp_relay_edit(args):
    """
    The guts for the following:

    netmisc (add|del) dhcp relay server <ipv4>
    netmisc (add|del) dhcp relay interface <interface>
    """
    command_ok = True

    if not os.path.exists(isc_dhcp_relay_scratchpad):
        if not os.path.exists(isc_dhcp_relay):
            print 'ERROR: %s does not exist, cannot configure DHCP relay' % isc_dhcp_relay
            return
        copyfile(isc_dhcp_relay, isc_dhcp_relay_scratchpad)

    servers = []
    interfaces = []
    lines = []

    with open(isc_dhcp_relay_scratchpad, 'r') as fh:
        for line in fh.readlines():
            line = line.strip()
            lines.append(line)

            if line.startswith('SERVERS'):
                re_key_value = re.search('^SERVERS\s*=\s*"(.*)"$', line)

                if re_key_value:
                    servers = re_key_value.group(1).strip().split()
                else:
                    print "ERROR: could not parse '%s'" % line

            elif line.startswith('INTF_CMD'):
                re_key_value = re.search('^INTF_CMD\s*=\s*"(.*)"$', line)

                if re_key_value:
                    intf_cmds_str = re_key_value.group(1).strip().replace(' -i ', ' ')

                    # Remove the -i from "-i swp1 -i swp2"
                    if intf_cmds_str.startswith('-i '):
                        intf_cmds_str = intf_cmds_str[3:].strip()

                    interfaces = intf_cmds_str.split()
                else:
                    print "ERROR: could not parse '%s'" % line

    if args.get('server'):
        ipv4 = args.get('<ipv4>')

        if args.get('add'):
            if ipv4 in servers:
                print 'DHCP relay SERVERS configuration already has %s' % ipv4
                command_ok = False
            else:
                servers.append(ipv4)

        elif args.get('del'):
            if ipv4 in servers:
                servers.remove(ipv4)
            else:
                print 'DHCP relay SERVERS configuration does not have %s' % ipv4
                command_ok = False

    elif args.get('interface'):
        ifname = args.get('<interface>')
        ifnames = []

        if ifname_is_glob(str(ifname)):
            for x in ifname_expand_glob(str(ifname)):
                ifnames.append(x)
        else:
            ifnames = [ifname, ]

        for ifname in ifnames:
            if args.get('add'):
                if ifname in interfaces:
                    print 'DHCP relay INTF_CMD configuration already has %s' % ifname
                    command_ok = False
                else:
                    interfaces.append(ifname)

            elif args.get('del'):
                if ifname in interfaces:
                    interfaces.remove(ifname)
                else:
                    print 'DHCP relay INTF_CMD configuration does not have %s' % ifname
                    command_ok = False

    servers = sort_for_humans(servers)
    interfaces = sort_for_humans(interfaces)
    # log.debug('servers: %s' % servers)
    # log.debug('interfaces: %s' % interfaces)
    # log.debug('lines: %s' % pformat(lines))

    with open(isc_dhcp_relay_scratchpad, 'w') as fh:
        for line in lines:
            if line.startswith('SERVERS'):
                fh.write('SERVERS="%s"\n' % ' '.join(servers))

            elif line.startswith('INTF_CMD'):
                if interfaces:
                    fh.write('INTF_CMD="-i %s"\n' % ' -i '.join(interfaces))
                else:
                    fh.write('INTF_CMD=""\n')

            else:
                fh.write(line + '\n')

    return command_ok


def hostname_valid(hostname):
    if len(hostname) < 1 or len(hostname) > 63:
        return False

    for x in hostname:
        if x.isalpha() or x.isdigit() or x == '-':
            pass
        else:
            return False

    return True


def hostname_edit(new_hostname):
    command_ok = True

    if not hostname_valid(new_hostname):
        print "ERROR: %s is an invalid hostname, hostnames may only contain letters, numbers, or -" % new_hostname
        return

    # execlist
    with open(execlist, 'w') as fh:
        fh.write('hostname %s\n' % new_hostname)
        fh.write('hostnamectl set-hostname %s\n' % new_hostname)
        fh.write('echo "You must re-login for your bash prompt to reflect the hostname change"\n')

    # /etc/hostname
    with open(etc_hostname_scratchpad, 'w') as fh:
        fh.write("%s\n" % new_hostname)

    old_hostname = check_output(['/bin/hostname']).strip()

    # /etc/hosts
    etc_hosts_lines = []
    with open(etc_hosts, 'r') as fh:
        for line in fh.readlines():
            line = line.strip()
            etc_hosts_lines.append(line)

    with open(etc_hosts_scratchpad, 'w') as fh:
        for line in etc_hosts_lines:
            re_old_hostname = re.search('^(.*\s+)%s$' % old_hostname, line)

            if re_old_hostname:
                fh.write("%s%s\n" % (re_old_hostname.group(1), new_hostname))
            else:
                fh.write(line + '\n')

    return command_ok


def build_doc_string():
    global __doc__

    # If netshow is present wrapper it
    if netshow_present:
        __doc__ = __doc__.replace('    netmisc show version',"""    netmisc show version

    # interface show commands
    netmisc clear counters
    netmisc show counters [json]
    netmisc show interface (all|bonds|bondmems) [mac|json]
    netmisc show interface [<interface>] [json]
    netmisc show lldp [json]
    netmisc show system""")

    if os.path.exists(isc_dhcp_relay):
        __doc__ = __doc__.replace('    netmisc show version', """    netmisc show version
    netmisc (add|del) dhcp relay server <ipv4>
    netmisc (add|del) dhcp relay interface <interface>""")

        __doc__ = __doc__.replace('show this screen and exit', """show this screen and exit
    access     : Summary of physical ports with l2 or l3 config
    bondmems   : Summary of bond members
    bonds      : Summary of bonds
    bridges    : Summary of ports with bridge members
    counters   : show netstat counters
    interface  : Summary info of all interfaces
    l3         : Summary of ports with an IP
    lldp       : Link Layer Discovery Protocol
    system     : System information
    trunks     : Summary of trunk interfaces
    all        : Everything
    json       : Print output in json
    mac        : Media Access Control
            """)

    # /etc/cumulus/ports.conf
    if os.path.exists(ports_conf):
        __doc__ = __doc__.replace('    netmisc show version',"""    netmisc show version

    # /etc/cumulus/ports.conf
    netmisc add interface <interface> breakout (4x10G|4x25G|2x50G|10G|40G|50G|100G|disabled)
            """)

        __doc__ = __doc__.replace('show this screen and exit', """show this screen and exit
    4x10G       : Split into four 10G interfaces
    4x25G       : Split into four 25G interfaces
    2x50G       : Split into two 50G interfaces
    10G         : 10G
    40G         : 40G
    50G         : 50G
    100G        : 100G
    disabled    : Disabled
    breakout    : Split an interface into multiple interface via a breakout cable
            """)


class ConfigWrapper(object):

    def __init__(self, show_linux_command, color_diffs):
        self.show_linux_command = show_linux_command
        self.color_diffs = color_diffs

    def get_managed_files(self):
        files = []
        files.append(isc_dhcp_relay)
        files.append(ports_conf)
        files.append(etc_hostname)
        files.append(etc_hosts)
        return files

    def get_service_cmd_for_file(self, filename):
        cmds = []

        if filename == isc_dhcp_relay:
            cmds.append(('/bin/systemctl', 'reset-failed', 'dhcrelay.service'))
            cmds.append(('/bin/systemctl', 'restart', 'dhcrelay.service'))

        elif filename == ports_conf:
            cmds.append(('/bin/systemctl', 'reset-failed', 'switchd.service'))
            cmds.append(('/bin/systemctl', 'restart', 'switchd.service'))

        elif filename == etc_hostname:
            new_hostname = None

            with open(etc_hostname, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()

                    if line:
                        new_hostname = line
                        break

            if new_hostname:
                cmds.append(('hostname', new_hostname))
                cmds.append(('hostnamectl', 'set-hostname', new_hostname))

        return cmds

    def abort_pending(self):
        if os.path.exists('/var/run/nclu/netmisc/'):
            for filename in os.listdir('/var/run/nclu/netmisc/'):
                full_filename = os.path.join('/var/run/nclu/netmisc/', filename)

                if os.path.isdir(full_filename):
                    rmtree(full_filename)
                else:
                    os.remove(full_filename)

    def commit_pending(self, verbose):
        files_modified = []

        # dhcp relay
        if os.path.exists(isc_dhcp_relay_scratchpad) and not files_match(isc_dhcp_relay, isc_dhcp_relay_scratchpad):
            shutil.move(isc_dhcp_relay_scratchpad, isc_dhcp_relay)
            files_modified.append(isc_dhcp_relay)

            for cmd in (['/bin/systemctl', 'enable', 'dhcrelay.service'],
                        ['/bin/systemctl', 'reset-failed', 'dhcrelay.service'],
                        ['/bin/systemctl', 'restart', 'dhcrelay.service']):

                if self.show_linux_command:
                    print "exec: %s" % ' '.join(cmd)

                try:
                    output = check_output(cmd, stderr=STDOUT)
                    if verbose:
                        print output
                except CalledProcessError as e:
                    print "\nWARNING: '%s' failed due to\n%s\n" % (' '.join(cmd), e)

                    try:
                        output = check_output(['/bin/systemctl', 'status', 'dhcrelay.service'], stderr=STDOUT)
                        print '\n/bin/systemctl status dhcrelay.service'
                        print '======================================'
                        print output
                    except CalledProcessError as e:
                        pass

        # hostname
        if os.path.exists(etc_hosts_scratchpad) and not files_match(etc_hosts, etc_hosts_scratchpad):
            shutil.move(etc_hosts_scratchpad, etc_hosts)
            files_modified.append(etc_hosts)

        if os.path.exists(etc_hostname_scratchpad) and not files_match(etc_hostname, etc_hostname_scratchpad):
            shutil.move(etc_hostname_scratchpad, etc_hostname)
            files_modified.append(etc_hostname)

        if os.path.exists(execlist):
            with open(execlist, 'r') as fh:
                for line in fh.readlines():
                    cmd = shlex.split(line.strip())

                    if self.show_linux_command:
                        print "exec: %s" % ' '.join(cmd)

                    try:
                        output = check_output(cmd, stderr=STDOUT)
                        if verbose:
                            print output
                    except CalledProcessError as e:
                        print "\nWARNING: '%s' failed due to\n%s\n" % (' '.join(cmd), e)

            os.remove(execlist)

        # /etc/cumulus/ports.conf
        if os.path.exists(ports_conf_scratchpad) and not files_match(ports_conf, ports_conf_scratchpad):
            shutil.move(ports_conf_scratchpad, ports_conf)
            files_modified.append(ports_conf)

            try:
                output = check_output(['/bin/systemctl', 'reset-failed', 'switchd'])
                if verbose:
                    print output

                output = check_output(['/bin/systemctl', 'restart', 'switchd'])
                if verbose:
                    print output

            except Exception:
                print "ERROR: Failed to restart switchd\n%s" % output

        return files_modified

    def get_configured_hostname(self):
        if os.path.exists(etc_hostname):
            with open(etc_hostname, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()

                    if line:
                        return line
        return None

    def get_configured_dhcp_relay(self):
        dhcp_servers = []
        dhcp_interfaces = []

        if os.path.exists(isc_dhcp_relay) and dhcp_relay_modified():
            with open(isc_dhcp_relay, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()

                    if line.startswith('SERVERS'):
                        re_key_value = re.search('^SERVERS\s*=\s*"(.*)"$', line)

                        if re_key_value:
                            dhcp_servers.extend(re_key_value.group(1).strip().split())

                    elif line.startswith('INTF_CMD'):
                        re_key_value = re.search('^INTF_CMD\s*=\s*"(.*)"$', line)

                        if re_key_value:
                            interfaces = re_key_value.group(1).replace('-i', '').strip().split()
                            if interfaces:
                                interfaces_as_glob = ifnames_to_glob(interfaces)

                                if interfaces_as_glob:
                                    dhcp_interfaces.append(interfaces_as_glob)
                                else:
                                    dhcp_interfaces.extend(interfaces)
            if dhcp_servers:
                dhcp_servers = sort_for_humans(dhcp_servers)

            if dhcp_interfaces:
                dhcp_interfaces = sort_for_humans(dhcp_interfaces)

        return (dhcp_servers, dhcp_interfaces)

    def get_ports_conf(self):
        ports = []

        if os.path.exists(ports_conf):
            with open(ports_conf, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()
                    re_line = re.search('^(\d+)\s*=\s*(\w+)$', line)

                    if re_line:
                        port_number = int(re_line.group(1))
                        port_speed = re_line.group(2)
                        ports.append((port_number, port_speed))

        return ports

    def show_pending(self, print_errors=True):
        diff_pending = False

        if os.path.exists(isc_dhcp_relay_scratchpad):
            if print_diff(isc_dhcp_relay, isc_dhcp_relay_scratchpad, self.color_diffs):
                diff_pending = True

        if os.path.exists(etc_hosts_scratchpad):
            if print_diff(etc_hosts, etc_hosts_scratchpad, self.color_diffs):
                diff_pending = True

        if os.path.exists(etc_hostname_scratchpad):
            if print_diff(etc_hostname, etc_hostname_scratchpad, self.color_diffs):
                diff_pending = True

        if os.path.exists(ports_conf_scratchpad):
            if print_diff(ports_conf, ports_conf_scratchpad, self.color_diffs):
                diff_pending = True
                print """\033[91m\n
********************************************************************
********************************************************************
NOTE: 'net commit' will run 'systemctl restart switchd' to apply the
change to /etc/cumulus/ports.conf. This can take a few seconds.
********************************************************************
********************************************************************
\033[0m
"""

        return diff_pending

    def show_config_files(self, user_may_edit):

        for filename in (etc_hostname,
                         etc_hosts,
                         isc_dhcp_relay,
                         ports_conf):
            if os.path.exists(filename) and os.path.getsize(filename) > 0:

                # If we haven't modified the default config file then do not display it
                if filename == isc_dhcp_relay and not dhcp_relay_modified():
                    continue

                print filename
                print '=' * len(filename)

                with open(filename, 'r') as fh:
                    print ''.join(fh.readlines())
                print ''

    def show_config_commands(self, user_may_edit):
        commands = []
        hostname = self.get_configured_hostname()

        if hostname:
            commands.append('net add hostname %s' % hostname)

        (dhcp_servers, dhcp_interfaces) = self.get_configured_dhcp_relay()

        for server in dhcp_servers:
            commands.append('net add dhcp relay server %s' % server)

        for interface in dhcp_interfaces:
            commands.append('net add dhcp relay interface %s' % interface)

        ports_by_speed = {}
        for (port_number, port_speed) in self.get_ports_conf():
            if port_speed not in ports_by_speed:
                ports_by_speed[port_speed] = []
            ports_by_speed[port_speed].append(port_number)

        for port_speed in sort_for_humans(ports_by_speed.keys()):
            port_numbers = ports_by_speed[port_speed]
            port_numbers_glob = numbers_to_glob(port_numbers)
            commands.append('net add port %s speed %s' % (port_numbers_glob, port_speed))

        return (commands, [])

    def show_config_summary(self, user_may_edit, summary):
        hostname = self.get_configured_hostname()

        if hostname:
            summary["hostname %s" % hostname] = []

        (dhcp_servers, dhcp_interfaces) = self.get_configured_dhcp_relay()
        if dhcp_servers or dhcp_interfaces:
            summary["dhcp relay"] = OrderedDict()

            if dhcp_servers:
                summary['dhcp relay']['servers'] = dhcp_servers

            if dhcp_interfaces:
                summary['dhcp relay']['interfaces'] = dhcp_interfaces

    def eval_command_line_args(self, cli):
        """
        NetworkDocopt will populate the self.args dictionary based on the command
        line args entered by the user
        """
        args = cli.args
        command_ok = True

        if args.get('show'):

            if args.get('hostname'):
                sys.stdout.write(check_output(('/bin/hostname',)))

            elif args.get('version'):
                with open('/etc/lsb-release', 'r') as fh:
                    for line in fh.readlines():
                        sys.stdout.write(line)

            # netshow
            else:
                vendor = vendor_check()
                show = None

                # 'access', 'bridges', 'bonds', 'bondmems', 'mgmt', 'l2', 'l3', 'phy', 'trunks', 'interface'
                if interface_related(cli):
                    if vendor == 'cumulus':
                        show = CumulusShowInterfaces(cli)
                    else:
                        show = ShowInterfaces(cli)

                elif cli.get('system'):
                    if vendor == 'cumulus':
                        show = CumulusShowSystem(cli)
                    else:
                        show = ShowSystem(cli)

                elif cli.get('lldp'):
                    if vendor == 'cumulus':
                        show = CumulusShowNeighbors(cli)
                    else:
                        show = ShowNeighbors(cli)

                elif cli.get('counters'):
                    cmd = ['/usr/cumulus/bin/cl-netstat',]

                    if args['json']:
                        cmd.append('-j')

                    try:
                        output = check_output(cmd, stderr=STDOUT)
                        print output
                    except Exception as e:
                        print "'%s' failed" % ' '.join(cmd)
                        print e
                    show = None

                if show:
                    print show.run()
                elif not cli.get('counters'):
                    print "ERROR: could not determine what data to show"

        elif args.get('clear'):

            if args.get('counters'):
                cmd = ['/usr/cumulus/bin/cl-netstat', '-c']
                try:
                    check_output(cmd, stderr=STDOUT)
                except Exception as e:
                    print "'%s' failed" % ' '.join(cmd)
                    print e

        # (add|del)
        else:
            if args.get('dhcp'):
                if args.get('relay'):
                    command_ok = dhcp_relay_edit(args)

            elif args.get('hostname'):
                command_ok = hostname_edit(args.get('<text>'))

            elif args.get('breakout'):
                command_ok = ports_conf_edit(args)

        return command_ok
